/*
 * alias.h: header for alias.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 *
 * @(#)$Id: alias.h,v 1.5 1994/07/02 02:38:10 mrg Exp $
 */

#ifndef _ALIAS_H_
#define _ALIAS_H_

#include "irc_std.h"

#define COMMAND_ALIAS 0
#define VAR_ALIAS 1
#define VAR_ALIAS_LOCAL 2

/*
 * The interfaces.
 * This has been divided up into "real" and "bogus" lists.
 */

/* REAL */
/*
 * This function is a general purpose interface to alias expansion.
 * The first argument is a dummy argument.
 * The second argument is the text to be expanded.
 * The third argument are the command line expandoes $0, $1, etc.
 * The fourth argument is a flag whether $0, $1, etc are used
 * The fifth argument, if set, controls whether only the first "command"
 *   will be expanded.  If set, this argument will be set to the "rest"
 *   of the commands (after the first semicolon, or the null).  If NULL,
 *   then the entire text will be expanded.
 */
extern	char *	expand_alias _((char *, char *, char *, int *, char **));

/*
 * This function is used to expand a named variable.
 * The argument is a NAME of an alias to expand (no $s!)
 * The name must not be a command line expando ($0, $1, etc)
 */
extern	char *	find_inline _((char *));


/*
 * This function allows you to add a global or local variable or cmdalias.
 * The first argument is one of the  *ALIAS* defines above.
 * The second argument is the name
 * The third argument is the stuff.
 */
extern	void	add_alias _((int, char *, char *));

/*
 * This function is the user interface to the /ALIAS command.
 * Noone should be calling it directly.
 */
extern	void	aliascmd _((char *, char *, char *));

/*
 * This is the user inteface to the /DUMP command.
 * Noone should be calling this function directly.
 */
extern	void	dumpcmd _((char *, char *, char *));

/*
 * This function is used to call a user-defined function.
 * Noone should be calling this directly except for call_function.
 */
extern	char *	call_user_function _((char *, char *));

/*
 * This function is used to see if a given alias name is an unambiguous
 * subset of a real alias.  If it is not, use match_alias to get all of
 * the ambiguous choices.
 *
 * This is a front end to find_alias.
 *
 * The first argument is the type of alias, (see above)
 * The second argument is the alias name
 * The third argument will be set to the number of matching aliases
 * The fourth argument will be set to the completed alias name, as long
 *    as it is the only nonambiguous completion.
 */
extern	char *	get_alias _((int, char *, int *, char **));

/*
 * This function is explicitly used for command completion.
 * The first argument is the alias name to complete
 * The second argument is set to the number of aliases expanded
 * The third argument is the type of alias to match (ugh.)
 * It returns a malloced double array of strings containing
 *   all of the aliases.
 */
extern	char ** match_alias _((char *, int *, int));

/*
 * This is the interface to the "expression parser"
 * The first argument is the expression to be parsed
 * The second argument is the command line expandoes ($0, $1, etc)
 * The third argument will be set if the command line expandoes are used.
 */
extern	char *	parse_inline _((char *, char *, int *));

/*
 * This function is sued to save all the current aliases to a global
 * file.  This is used by /SAVE and /ABORT.
 */
extern	void	save_aliases _((FILE *, int));

/*
 * This function is in functions.c
 * This function allows you to execute a primitive "BUILT IN" expando.
 * These are the $A, $B, $C, etc expandoes.
 * The argument is the character of the expando (eg, 'A', 'B', etc)
 *
 * This is in functions.c
 */
extern char *built_in_alias(char);



/* BOGUS */

/*
 * This function is used by parse_command to directly execute an alias.
 * Noone should be calling this function directly. (call parse_line.)
 */
extern	void	execute_alias _((char *, char *, char *));

/*
 * This is in functions.c
 * This is only used by next_unit and expand_alias to call built in functions.
 * Noone should call this function directly.
 */
extern char *call_function(char *, char *, char *, int *);



/*
 * These are the two primitives for runtime stacks.
 */
extern	void	make_local_stack _((char *));
extern	void	destroy_local_stack _((void));
extern	void	set_current_command _((char *));
extern	void	bless_local_stack _((void));
extern	void	unset_current_command _((void));
extern	void	dump_call_stack	_((void));
extern	void	panic_dump_call_stack _((void));

/*
 * This is the alias interface to the /STACK command.
 */
extern	void	do_stack_alias _((int, char *, int));

#endif /* _ALIAS_H_ */
